//==========================================================================
//
//      plf_misc.c
//
//      HAL platform miscellaneous functions
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    msieweke
// Contributors: nickg, jlarmour, dmoseley, jskov
// Date:         2003-06-16
// Purpose:      HAL miscellaneous functions
// Description:  This file contains miscellaneous functions provided by the
//               HAL.
//
//####DESCRIPTIONEND####
//
//========================================================================*/

#define CYGARC_HAL_COMMON_EXPORT_CPU_MACROS
#include <pkgconf/hal.h>

#include <cyg/infra/cyg_type.h>         // Base types
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros
#include <cyg/infra/diag.h>             // diag output

#include <cyg/hal/hal_arch.h>           // architectural definitions
#include <cyg/hal/hal_intr.h>           // Interrupt handling
#include <cyg/hal/hal_cache.h>          // Cache handling
#include <cyg/hal/hal_if.h>
#include <cyg/hal/hal_diag.h>
#include <cyg/hal/bcm53xx_regs.h>

#include <cyg/kernel/kapi.h>


//--------------------------------------------------------------------------

cyg_uint32 *ChipCcore   = (cyg_uint32*)( 0xb8000000 );
cyg_uint32 *MIPScore    = (cyg_uint32*)( 0xb8005000 );
int         MIPScoreNum = 5;

// If the application wants a periodic timer, it can override this value.
cyg_uint32  MIPSTimerReloadValue __attribute__ ((weak)) = 0;


//--------------------------------------------------------------------------

externC void diag_write_string( const char* );
static  void hal_init_interrupts( void );


//--------------------------------------------------------------------------

void hal_platform_init(void)
{
    cyg_uint32 *Core = (cyg_uint32*)( 0xb8000000 );
    int         i;
    cyg_uint16  CoreType;

    // Set up eCos/ROM interfaces
    hal_if_init();

    // Find the MIPS core.  We only check cores 0-5 because we're pretty sure
    // they exist.  If we try to check a core which doesn't exist, we get a
    // bus error.
    for ( i = 0; i < 6; i++ )
    {
        CoreType = (( Core[SB_IDHigh] >> 4 ) & 0xfff );
        if (( CoreType == 0x805 ) ||
            ( CoreType == 0x816 ) ||
            ( CoreType == 0xfff ))
        {
            break;
        }
        else
        {
            Core += (0x1000 / 4);
        }
    }
    CoreType = (( Core[SB_IDHigh] >> 4 ) & 0xfff );
    if (( CoreType == 0x805 ) ||
        ( CoreType == 0x816 ))
    {
        MIPScore    = Core;
        MIPScoreNum = i;
    }

#if defined(CYGPKG_CYGMON)
    diag_write_string(" CYGMON");
#elif defined(CYGPKG_REDBOOT)
    diag_write_string(" RedBoot");
#else
    diag_write_string(" eCos");
#endif

    hal_init_interrupts();

    // Load the timer and enable the timer interrupt.
    MIPScore[MIPS_Timer0]       = MIPSTimerReloadValue;
    MIPScore[SB_IntEnableMask]  = 1;
}


//------------------------------------------------------------------------
// In a 53xx each core can generate an interrupt, and each core interrupt can
// be routed to a MIPS external interrupt (1-4).  Also, a group of interrupts
// can be routed to MIPS interrupt 0.  We'll route each interesting interrupt-
// generating core to an interrupt, like this:
// core 0 = int 0               ChipCommon (UART)
// core 1 = int 1               Ethernet MAC
// core 2 = int 2               IPSec
// core 3 = int 3               USB MAC
// core 4 = int 4               PCI
// core 5 = int 0               MIPS (timer)
// core 6 --- no interrupt      Memory controller

// This means cores 0 and 5 both use ISR 0, which is a problem because we
// have no way to test whether either core's interrupt is active.  To find
// which is active, we have to read core registers, which we would prefer not 
// to do.  And there's no register to tell us when the UART interrupt is
// active, so we have to call its ISR every time for interrupt 0.

#define NUM_INTERRUPTS 5
#define NUM_ISRS       6

static void           (*uartIsr)( void *data ) = { NULL };
static void           *uartIsrData             = { NULL };

static void           (*subIsr[NUM_ISRS])( void *data ) = { NULL };
static void           *subIsrData[NUM_ISRS]             = { NULL };

static cyg_handle_t   hal_isr_handle[NUM_INTERRUPTS];
static cyg_interrupt  hal_isr_interrupt[NUM_INTERRUPTS];


//------------------------------------------------------------------------
// The main ISR is pretty simple - disable the interrupt and let eCos run
// the DSR.
// We've added a feature to automatically reload the MIPS timer.  This gives
// us a periodic timer more like the ones in 33xx cores.
static cyg_uint32 hal_main_isr( cyg_vector_t    vector, 
                                cyg_addrword_t  data)
{
    // Disable the interrupt until after the DSR runs.
    HAL_INTERRUPT_MASK( vector );

    // Interrupt 0 is for the UART and MIPS timer only.  If we've set up
    // an auto-reload value for the MIPS timer, see if we need to load it.
    if ((vector == 0) && (MIPSTimerReloadValue != 0))
    {
        // If the MIPS Timer is active
        if ((MIPScore[SB_IntStatus] & MIPScore[SB_IntEnableMask]) != 0)
        {
            // Reload the timer (simulates a periodic timer).
            MIPScore[MIPS_Timer0]   = MIPSTimerReloadValue;
        }
    }

    // This ISR's only function is to kick off the DSR.
    return( CYG_ISR_HANDLED | CYG_ISR_CALL_DSR );
}


//------------------------------------------------------------------------
static void hal_main_dsr( cyg_vector_t    vector, 
                          cyg_ucount32    count,
                          cyg_addrword_t  data )
{
    // Acknowledge the interrupt, because the external "ISR" may unmask it.
    HAL_INTERRUPT_ACKNOWLEDGE( vector );

    // Call the associated interrupt routine.  The vector number will
    // be 0-4.  Vector 0 is a special case because both ChipCommon and MIPS
    // route through vector 0.  And there's no way to tell if the UART
    // interrupt is active, so we have to call it always.
    if (vector == 0)
    {
        // First check the MIPS ISR.
        if (( MIPScore[SB_IntStatus] & MIPScore[SB_IntEnableMask] ) != 0 )
        {
            if (subIsr[5] != NULL)
            {
                subIsr[5]( subIsrData[5] );
            }
            // Clear the timer interrupt.
            MIPScore[SB_IntStatus]  = 1;
        }
        if (( ChipCcore[SB_IntStatus] & ChipCcore[SB_IntEnableMask] ) != 0 )
        {
            if (subIsr[0] != NULL)
            {
                subIsr[0]( subIsrData[0] );
            }
        }
        if (uartIsr != NULL)
        {
            uartIsr( uartIsrData );
        }
    }
    else 
    {
        if (subIsr[vector] != NULL)
        {
            subIsr[vector]( subIsrData[vector] );
        }
    }
//    hal_diag_write_char( '0' + vector );
}


//------------------------------------------------------------------------
// Configure our ISRs for interrupts 0-4.

static void hal_init_interrupts()
{
    int         intVec;

    // Route only the core 0 (ChipCommon) and MIPS interrupts to vector 0.
    MIPScore[SB_IntVec]  = ( 1 << 0 ) | ( 1 << MIPScoreNum );

    // Route interrupts from cores 4321 to interrupts 4321.  This may be over-
    // ridden in the application, in case it needs the interrupt for a core
    // numbered higher than 4.
    MIPScore[SB_IPSFlag] = 0x04030201;

    for ( intVec = 0; intVec < NUM_INTERRUPTS; intVec++ )
    {
        cyg_interrupt_create( intVec,               // MIPS interrupt number
                              0,                    // priority (unused in MIPS)
                              0,                    // Data item passed to ISR
                              &hal_main_isr,
                              &hal_main_dsr,
                              &hal_isr_handle[intVec],
                              &hal_isr_interrupt[intVec] );
        cyg_interrupt_attach( hal_isr_handle[intVec] );
        HAL_INTERRUPT_UNMASK( intVec );
    }
}


//------------------------------------------------------------------------
// Find the interrupt mapped to a particular core.
static int hal_convert_core_to_vector( cyg_uint32 coreNum )
{
    cyg_uint32  IPSFlag;
    int         vector;

    // The UART and MIPS interrupts are special cases.  They're routed to vector 0.
    // So is core 0 (ChipCommon).
    if (( coreNum == UART0_SUB_IRQ ) ||
        ( coreNum == MIPScoreNum )   ||
        ( coreNum == 0 ))
    {
        return 0;
    }

    // The IPSFlag register indicates core interrupt routing for vector 4,3,2,1.
    IPSFlag = MIPScore[SB_IPSFlag];

    // Look at the interrupt routing, from right to left.  If one matches the
    // requested core interrupt, return the vector.
    for ( vector = 1; vector <= 4; vector++ )
    {
        if (( IPSFlag & 0xff ) == coreNum )
        {
            return vector;
        }
        IPSFlag >>= 8;
    }
    return -1;
}


//------------------------------------------------------------------------
// Register an interrupt handler.

bool hal_map_interrupt( cyg_uint32  coreNum, 
                        void      (*newIsr)(void *), 
                        void       *newData )
{
    int  vector;

    // The UART is a special case.  It comes through core 0, but it doesn't
    // show up in that core's interrupt status register.
    if (coreNum == UART0_SUB_IRQ )
    {
        uartIsr     = newIsr;
        uartIsrData = newData;
        return true;
    }

    // The MIPS core is a special case.  Its timer interrupt is routed
    // through vector 0, but we keep its ISR in slot 5.
    if ( coreNum == MIPScoreNum )
    {
        subIsr[5]     = newIsr;
        subIsrData[5] = newData;
        return true;
    }

    // We only support interrupts for cores 0-5, but the UART interrupt is
    // treated as core 7.  Core 6 is the memory controller, and we don't
    // handle any interrupts for it.
    if (( vector = hal_convert_core_to_vector( coreNum )) != -1 )
    {
        subIsr[vector]     = newIsr;
        subIsrData[vector] = newData;
        return true;
    }
    else
    {
        return false;
    }
}


//------------------------------------------------------------------------
// Enable an interrupt.

void hal_enable_interrupt( cyg_uint32 coreNum )
{
    int vector;

    // Find the vector which receives the core's interrupt.
    vector = hal_convert_core_to_vector( coreNum );
    if ( vector != -1 )
    {
        // The MIPS core is a special case.  Its timer interrupt is routed
        // through vector 0.  We never want to turn off interrupt 0 because
        // it's used for the UART.
        if ( coreNum == MIPScoreNum )
        {
            MIPScore[SB_IntEnableMask] = 1;
        }
        else
        {
            HAL_INTERRUPT_UNMASK( vector );
        }
    }
}


//------------------------------------------------------------------------
// Disable an interrupt.

void hal_disable_interrupt( cyg_uint32 coreNum )
{
    int vector;

    // Find the vector which receives the core's interrupt.
    vector = hal_convert_core_to_vector( coreNum );
    if ( vector != -1 )
    {
        // The MIPS core is a special case.  Its timer interrupt is routed
        // through vector 0.  We never want to turn off interrupt 0 because
        // it's also used for the UART.
        if ( coreNum == MIPScoreNum )
        {
            MIPScore[SB_IntEnableMask] = 0;
        }
        else
        {
            HAL_INTERRUPT_MASK( vector );
        }
    }
}


/*------------------------------------------------------------------------*/
/* Reset support                                                          */

void hal_bcm53xx_reset(void)
{
    diag_write_string( "(hal_bcm53xx_reset) Resetting...\n" );
//***** unfinished
    for(;;);                            // wait for it
}


typedef cyg_uint32 uint32;
typedef cyg_uint16 uint16;
typedef cyg_uint8  uint8;

//*****************************************************************************
// Calculate RAM size by writing test values to RAM.  This counts on the address
// being partially decoded, so RAM will be duplicated (RAM space)/(true size) times.
//
// DANGER: Since this function modifies memory owned by other tasks, it is
// not interrupt safe.  To guard against memory corruption, interrupts are
// disabled while it runs.
//
// We shouldn't have to worry about overwriting the stack and local variables,
// because the initial stack grows downward from location 0x10000.

static uint32 CalculateRamSize(void)
{
#if 0
    volatile SdramControllerRegs  *SDRAM = (SdramControllerRegs*) SDRAM_BASE;
#endif
    unsigned long __state;

    HAL_DISABLE_INTERRUPTS(__state);

    uint32   DetectedRamSize = 0; // Set to 0 to eliminate compile warning.
    uint32   SavedValues[8];      // Saved values for 1M - 128M

    int               i;
    int               ProbeLimit;
    volatile uint32  *ProbeAddress;
    uint32            RamBase        = 0xa0000000; // RAM address 0 in non-cached space
    uint32            RamBaseValue;

#if 0
    // These are the base-2 logs of address space size (in megabytes) for values
    // in the memory base register/size field.
    static uint8  MemSizeValues[8] = { 1, 3, 4, 5, 6, 7, 8, 9 };

    // Size of the memory address space - where 0=2M, 1=8M, 2=16M, 3=32M, 4=64M...
    uint32   MemSizeReg;

    // The memory-base-register/memory-space says how large RAM _could_ be.  We 
    // can't probe outside this address space without getting a bus error.
    MemSizeReg  = SDRAM->memoryBase & SD_MEMSIZE_MASK;

    // Convert from the register value to the base-2 logarithm of the memory 
    // address space size in megabytes.
    ProbeLimit   = MemSizeValues[MemSizeReg];
#else
    // Limit memory probing to 16 Mbytes.
    ProbeLimit   = 4;
#endif

    // Save the value at RAM location 0.
    ProbeAddress  = (uint32*) RamBase;
    RamBaseValue  = *ProbeAddress;

    // Save values at address 1M, 2M, 4M, 8M, etc.  Then store 0 in each location.
    // ( Remember that 1<<20 is 1M. )
    for (i = 0; i < ProbeLimit; i++)
    {
        ProbeAddress   = (uint32*) (RamBase + (1 << (i + 20)));
        SavedValues[i] = *ProbeAddress;
        *ProbeAddress  = 0;
    }

    #define PROBE_VALUE 0xf0f0f0f0

    // Write a non-zero value to RAM location 0.  Where RAM wraps around, this
    // value will show up above the top of memory.
    ProbeAddress  = (uint32*) RamBase;
    *ProbeAddress = PROBE_VALUE;
    
    // Start probing at memory address 1Meg.  Technically it's not necessary to
    // probe addresses at 1M or 4M, but this way is simpler.
    for (i = 0; i < ProbeLimit; i++)
    {
        ProbeAddress  = (uint32*) (RamBase + (1 << (i + 20)));
        if (*ProbeAddress == PROBE_VALUE)
        {
            break;
        }
        // Restore the saved value for this address.
        *ProbeAddress = SavedValues[i];
    }
    // RAM size is 2^i megabytes...
    DetectedRamSize = 1 << (i + 20);

    // Restore the value we found at RAM address 0.
    ProbeAddress  = (uint32*) RamBase;
    *ProbeAddress =  RamBaseValue;

    HAL_RESTORE_INTERRUPTS(__state);

    return DetectedRamSize;
}


//------------------------------------------------------------------------
// Run-time memory size info.

// Allow the application to set an arbitrary limit on memory usage.  If the
// application defines this variable, that's used as the size limit.
cyg_uint32 hal_mips_mem_limit __attribute__ ((weak)) = 64*1024*1024;

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

cyg_uint8 *hal_mips_mem_real_region_top( cyg_uint8 *_regionend_ )
{
  
    static cyg_uint8  *memTop   = NULL;
    // Don't assume which segment to put the "memory top" value.  Use the
    // same segment as the one passed in.
    cyg_uint32         segMask  = (cyg_uint32)_regionend_ & 0xf0000000;

    if (memTop == NULL)
    {
        // We reserve 64k at the top of RAM, apparently for the vxWorks boot
        // loader, which we aren't using.
        memTop = (cyg_uint8 *)(MIN( CalculateRamSize(), hal_mips_mem_limit ) - (64*1024) + segMask);
    }
    return memTop;
}

/*------------------------------------------------------------------------*/
/* End of plf_misc.c                                                      */
