//=============================================================================
//
//      pcmb_serial.c
//
//      HAL diagnostic output code
//
//=============================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):   proven
// Contributors:proven
// Date:        1998-10-05
// Purpose:     HAL diagnostic output
// Description: Implementations of HAL diagnostic output support.
//
//####DESCRIPTIONEND####
//
//=============================================================================

#include <pkgconf/hal.h>

#include <cyg/infra/cyg_type.h>         // base types

#include <cyg/hal/hal_arch.h>           // basic machine info
#include <cyg/hal/hal_intr.h>           // interrupt macros
#include <cyg/hal/hal_io.h>             // IO macros
#include <cyg/hal/drv_api.h>
#include <cyg/hal/hal_if.h>             // interface API
#include <cyg/hal/hal_misc.h>

// The 16550 registers are defined here.
#include <cyg/hal/bcm53xx_regs.h>


/*---------------------------------------------------------------------------*/
/* From serial_16550.h */
#define CYG_DEVICE_SERIAL_BAUD_MSB        0x00
#if CYGNUM_HAL_VIRTUAL_VECTOR_CONSOLE_CHANNEL_BAUD==9600
    #define CYG_DEVICE_SERIAL_BAUD_LSB        0x0C
#endif
#if CYGNUM_HAL_VIRTUAL_VECTOR_CONSOLE_CHANNEL_BAUD==19200
    #define CYG_DEVICE_SERIAL_BAUD_LSB        0x06
#endif
#if CYGNUM_HAL_VIRTUAL_VECTOR_CONSOLE_CHANNEL_BAUD==38400
    #define CYG_DEVICE_SERIAL_BAUD_LSB        0x03
#endif
#if CYGNUM_HAL_VIRTUAL_VECTOR_CONSOLE_CHANNEL_BAUD==57600
    #define CYG_DEVICE_SERIAL_BAUD_LSB        0x02
#endif
#if CYGNUM_HAL_VIRTUAL_VECTOR_CONSOLE_CHANNEL_BAUD==115200
    #define CYG_DEVICE_SERIAL_BAUD_LSB        0x01
#endif

#ifndef CYG_DEVICE_SERIAL_BAUD_LSB
    #error Missing/incorrect serial baud rate defined - CDL error?
#endif


//-----------------------------------------------------------------------------

typedef struct {
    cyg_uint8 *base;
    cyg_int32  msec_timeout;
    int        isr_vector;
} channel_data_t;

static channel_data_t channels[1] = {
    { (cyg_uint8*)BCM_SER_53XX_BASE, 1000, 0}
};


//=============================================================================
// Basic channel functions

static void
cyg_hal_plf_serial_init_channel(void* __ch_data)
{
    volatile cyg_uint8 *UART;

    // Some of the diagnostic print code calls through here with no idea what the ch_data is.
    // Go ahead and assume it is channels[0].
    if (__ch_data == 0)
        __ch_data = (void*)&channels[0];

    UART = ((channel_data_t*)__ch_data)->base;

    UART[IER] = 0;   
    UART[IER] = 0;
    UART[MCR] = 0;    

    UART[LCR] = LCR_DLAB;
    // Leave the values from the bootloader.
//    UART[DLL] = CYG_DEVICE_SERIAL_BAUD_LSB;
//    UART[DLM] = CYG_DEVICE_SERIAL_BAUD_MSB;
    // 8-1-no parity.
    UART[LCR] = LCR_8BIT;

    UART[FCR] = 0x07;  // Enable & clear FIFO, in case the bootloader didn't.
}

//-----------------------------------------------------------------------------

void
cyg_hal_plf_serial_putc(void *__ch_data, char c)
{
    volatile cyg_uint8 *UART;
    cyg_uint8           OkToSend = LSR_THRE;

    // Some of the diagnostic print code calls through here with no idea what the ch_data is.
    // Go ahead and assume it is channels[0].
    if (__ch_data == 0)
        __ch_data = (void*)&channels[0];
    
    UART = ((channel_data_t*)__ch_data)->base;

    // If IER_TXFIFO is set, the sense of LSR_THRE is reversed.
    if ( UART[IER] & IER_TXFIFO )
    {
        OkToSend = 0;
    }
    while (( UART[LSR] & LSR_THRE ) != OkToSend )
    {
    }

    UART[THR] = c;

}

//-----------------------------------------------------------------------------

static cyg_bool
cyg_hal_plf_serial_getc_nonblock(void* __ch_data, cyg_uint8* ch)
{
    volatile cyg_uint8 *UART;
    cyg_uint8           lsr;

    // Some of the diagnostic print code calls through here with no idea what the ch_data is.
    // Go ahead and assume it is channels[0].
    if (__ch_data == 0)
        __ch_data = (void*)&channels[0];
    
    UART = ((channel_data_t*)__ch_data)->base;

    lsr = UART[LSR];
    if ( (lsr & LSR_DR) == 0 )
        return false;

    *ch = UART[RBR];

    return true;
}

//-----------------------------------------------------------------------------

cyg_uint8
cyg_hal_plf_serial_getc(void* __ch_data)
{
    cyg_uint8 ch;

    while ( !cyg_hal_plf_serial_getc_nonblock(__ch_data, &ch) )
        ; // Do nothing

    return ch;
}

//=============================================================================
// Call IF support

static void
cyg_hal_plf_serial_write(void* __ch_data, const cyg_uint8* __buf, 
                         cyg_uint32 __len)
{
    while ( __len-- > 0 )
        cyg_hal_plf_serial_putc(__ch_data, *__buf++);
}

static void
cyg_hal_plf_serial_read(void* __ch_data, cyg_uint8* __buf, cyg_uint32 __len)
{
    while ( __len-- > 0 )
        *__buf++ = cyg_hal_plf_serial_getc(__ch_data);
}

static cyg_bool
cyg_hal_plf_serial_getc_timeout(void* __ch_data, cyg_uint8* ch)
{
    int             delay_count;
    channel_data_t *chan;
    cyg_bool        res;

    // Some of the diagnostic print code calls through here with no idea what the ch_data is.
    // Go ahead and assume it is channels[0].
    if (__ch_data == 0)
        __ch_data = (void*)&channels[0];

    chan = (channel_data_t*)__ch_data;

    delay_count = chan->msec_timeout * 10; // delay in .1 ms steps

    for ( ;; )
    {
        res = cyg_hal_plf_serial_getc_nonblock(__ch_data, ch);
        if ( res || (delay_count-- == 0) )
            break;

        CYGACC_CALL_IF_DELAY_US(100);
    }

    return res;
}


static int
cyg_hal_plf_serial_control(void *__ch_data, __comm_control_cmd_t __func, ...)
{
    static int               irq_state  = 0;
    int                      ret        = 0;
    channel_data_t          *chan;
    volatile cyg_uint8      *UART;

    // Some of the diagnostic print code calls through here with no idea what the ch_data is.
    // Go ahead and assume it is channels[0].
    if (__ch_data == 0)
        __ch_data = (void*)&channels[0];

    chan =  (channel_data_t*)__ch_data;
    UART = ((channel_data_t*)__ch_data)->base;

    switch ( __func )
    {
        case __COMMCTL_IRQ_ENABLE:
            irq_state = 1;

            UART[IER] = IER_RCV;
            UART[MCR] = MCR_INT|MCR_DTR|MCR_RTS;

            HAL_INTERRUPT_UNMASK(chan->isr_vector);
            break;
        case __COMMCTL_IRQ_DISABLE:
            ret = irq_state;
            irq_state = 0;

            UART[IER] = 0;

            HAL_INTERRUPT_MASK(chan->isr_vector);
            break;
        case __COMMCTL_DBG_ISR_VECTOR:
            ret = chan->isr_vector;
            break;
        case __COMMCTL_SET_TIMEOUT:
            {
                va_list ap;

                va_start(ap, __func);

                ret = chan->msec_timeout;
                chan->msec_timeout = va_arg(ap, cyg_uint32);

                va_end(ap);

                break;
            }        
        default:
            break;
    }


    return ret;
}

static int
cyg_hal_plf_serial_isr(void         *__ch_data, 
                       int          *__ctrlc, 
                       CYG_ADDRWORD  __vector, 
                       CYG_ADDRWORD  __data)
{
    int                  res     = 0;
    char                 c;
    cyg_uint8            lsr;
    channel_data_t      *chan;
    volatile cyg_uint8  *UART;

    // Some of the diagnostic print code calls through here with no idea what the ch_data is.
    // Go ahead and assume it is channels[0].
    if (__ch_data == 0)
        __ch_data = (void*)&channels[0];

    chan =  (channel_data_t*)__ch_data;
    UART = ((channel_data_t*)__ch_data)->base;

    cyg_drv_interrupt_acknowledge(chan->isr_vector);

    *__ctrlc = 0;
    lsr = UART[LSR];
    if ( (lsr & LSR_DR) != 0 )
    {

        c = UART[RBR];
        if ( cyg_hal_is_break( &c , 1 ) )
            *__ctrlc = 1;

    }

    res = CYG_ISR_HANDLED;

    return res;
}

void cyg_hal_plf_serial_init(void)
{
    hal_virtual_comm_table_t   *comm;

    int cur   = CYGACC_CALL_IF_SET_CONSOLE_COMM(CYGNUM_CALL_IF_SET_COMM_ID_QUERY_CURRENT);

    // Disable interrupts.
    HAL_INTERRUPT_MASK(channels[0].isr_vector);

    // Init
    cyg_hal_plf_serial_init_channel((void *)&channels[0]);

    // Setup procs in the vector table

    // Set channel 0
    CYGACC_CALL_IF_SET_CONSOLE_COMM(0);
    comm = CYGACC_CALL_IF_CONSOLE_PROCS();
    CYGACC_COMM_IF_CH_DATA_SET(*comm, &channels[0]);
    CYGACC_COMM_IF_WRITE_SET(*comm, cyg_hal_plf_serial_write);
    CYGACC_COMM_IF_READ_SET(*comm, cyg_hal_plf_serial_read);
    CYGACC_COMM_IF_PUTC_SET(*comm, cyg_hal_plf_serial_putc);
    CYGACC_COMM_IF_GETC_SET(*comm, cyg_hal_plf_serial_getc);
    CYGACC_COMM_IF_CONTROL_SET(*comm, cyg_hal_plf_serial_control);
    CYGACC_COMM_IF_DBG_ISR_SET(*comm, cyg_hal_plf_serial_isr);
    CYGACC_COMM_IF_GETC_TIMEOUT_SET(*comm, cyg_hal_plf_serial_getc_timeout);

    // Restore original console
    CYGACC_CALL_IF_SET_CONSOLE_COMM(cur);
}

void
cyg_hal_plf_comms_init(void)
{
    static int initialized = 0;

    if (initialized)
        return;

    initialized = 1;

    cyg_hal_plf_serial_init();
}

//-----------------------------------------------------------------------------
// End of pcmb_serial.c
